"use strict";


const crypto = require('crypto');

const Cron = require('cron');
const CronJob = Cron.CronJob;
const CodeModule = require('../helper/module');
const Db = require('../helper/db');
const args = require('./args');

const shortInterval = 5 * 60 * 1000; //5 minutes
const longInterval = 24 * 60 * 60 * 1000; //24 hours
let interval;

const STATUS = {
  EXECUTING: 'executing',
  SUCCESS: 'ok',
  ERROR: 'failed',
  ABORTED: 'aborted'
};

const cronjobs = {}; // map: job ID --> cron job object

let maintenanceRoutine;//disable maintenance with clearInterval(maintenanceRoutine);
let activeJobs = 0; //the number of jobs active on this node

const range = (startInclusive, endExclusive) => (Array.from(new Array(endExclusive), (x, i) => i + startInclusive));
const RANGES = [
  range(0, 60),
  range(0, 60),
  range(0, 24),
  range(1, 32),
  range(0, 12),
  range(0, 7)
];

const handleJob = (job, type) => {
  if (type === 'insert' || type === 'update') {
    // node can always read and write jobs:
    job.acl.allowReadAccess('/db/Role/2');
    job.acl.allowWriteAccess('/db/Role/2');

    //stop job, if already running:
    let cronjob = cronjobs[job.id];
    if (cronjob) {
      cronjob.stop();
    }

    if (!job.startsAt) {
      job.startsAt = new Date(0);
    }

    try {
      cronjob = schedule(job);
      job.nextExecution = estimateNextExecutionTime(job, cronjob);
    } catch (e) {
      throw new Abort('Invalid cron job parameters: ' + (e.stack || e.message), job.toJSON());
    }

    if (activeJobs === 1) {
      scheduleMaintenance(true);
    }
  }
};

const implementHashes = (pattern, name) => {
  if (pattern) {
    pattern = pattern.replace(/\s+/g, " "); //remove double white spaces and so on
    if (pattern.indexOf('H') !== -1) {
      const hexHash = crypto.createHash('md5').update(name || '').digest('hex');
      const hash = parseInt(hexHash, 16);

      const hashedPattern = pattern.split(" ");

      for (let i = 0; i < hashedPattern.length; i++) {
        if (hashedPattern[i] === "H") {
          const range = RANGES[i];
          const bucket = hash % range.length;
          hashedPattern[i] = `${range[bucket]}`;
        }
      }
      return hashedPattern.join(" ");
    }
  }
  return pattern;
};

const schedule = (job, runImmediately) => {
  if (cronjobs[job.id])
    return cronjobs[job.id];

  //use a function statement here to access the cronjob object through the "this" context
  function failsafeTodo() {
    const now = Date.now();
    const cronjob = this;
    const nextTime = estimateNextExecutionTime(job, cronjob);

    if (job.startsAt.getTime() - now > 0) {// job starts later --> skip execution!
      return;
    }

    if (job.expiresAt && job.expiresAt.getTime() - now < 1) {// job is expired --> cancel!
      cronjob.stop();
      return;
    }

    const localDb = Db.createNodeDBForJob();

    log('executing job: ' + job);
    let status;
    claimExecution(localDb, job, nextTime).then(updatedJob => {
      if (!updatedJob) {// job was deleted or is handled by another node
        log('Job could not be claimed for execution --> canceling locally: ' + job);
        cronjob.stop();
        return;
      }

      job = updatedJob;

      status = newStatus(localDb, job);
      status.status = STATUS.EXECUTING;
      return status.save().then(status => {
        const module = CodeModule.require('/code/' + job.module, {});
        const code = module.run || module.call;
        if (!code) {
          log('Neither call nor run function defined in module used by job: ' + job, 'warn');
          throw new Abort('Neither call nor run function defined in code module.');
        }
        return code.call(module, localDb, status, job);
      }).then(() => {
        log('successfully executed job: ' + job, 'debug');
        status.status = STATUS.SUCCESS;
      }).catch(err => {
        status.status = err instanceof Abort ? STATUS.ABORTED : STATUS.ERROR;
        status.error = err && err.stack;
        log('error while executing job: ' + job, 'warn');
      }).then(() => {
        return status.save();
      }).then(() => {
        //reflect user changes on the job object
        return job.save();
      });
    });
  };

  let cleanup = () => {
    if (cronjobs[job.id]) {
      activeJobs--;
      delete cronjobs[job.id];
    }
  };

  try {
    job.cronpattern = implementHashes(job.cronpattern, job.module);
    const cronjob = new CronJob({
      cronTime: job.cronpattern,
      onTick: failsafeTodo,
      onComplete: cleanup,
      start: true,
      //timeZone: 'America/Los_Angeles',
      runOnInit: !!runImmediately
    });

    cronjobs[job.id] = cronjob;
    activeJobs++;

    log('scheduling job (' + activeJobs + ' active jobs in total): ' + job);
    return cronjob;
  } catch (err) {
    cleanup();
    throw err;
  }
};

const estimateNextExecutionTime = (job, cronjob) => {
  // CronTime return moment.js dates
  const nextExecution = cronjob.cronTime.sendAt().toDate();
  return job.expiresAt && nextExecution.getTime() > job.expiresAt.getTime()? null: nextExecution;
};

const claimExecution = (localDb, currentJob, nextTime) => {
  return localDb['jobs.Definition'].load(currentJob.id).then(job => {
    if (!job || job.nextExecution.getTime() > Date.now()) {
      return null; // job has been deleted or claimed by other node
    }

    job.nextExecution = nextTime;

    return job.save(); // claim job
  }).catch(err => {
    return null;
  });
};

const findAllUnexpiredJobs = () => {
  let query = Db.getGlobalDB()['jobs.Definition'].find();
  return query.or(
      query.greaterThan("expiresAt", new Date()),
      query.isNull("expiresAt")
  );
};

const checkForJobs = () => {
  if (activeJobs > 0) {
    return Promise.resolve(true);
  } else {
    return findAllUnexpiredJobs().count(count => count > 0);
  }
};

const scheduleMaintenance = (areThereActiveJobs) => {
  clearTimeout(maintenanceRoutine);

  //reschedule:
  // - high-frequency check when there are jobs
  // - check every 24 hours when there are no jobs
  interval = areThereActiveJobs ? shortInterval : longInterval;

  log('scheduling next maintenance in (' + (interval / (60000)) + ' minutes (' + (interval / 3600000) + ' hours)');
  maintenanceRoutine = setTimeout(maintenance.bind(this, false), interval);
};

const maintenance = (initialRun) => {
  Promise.resolve().then(() => {
    log('Maintenance: checking for overdue jobs...');
    let acceptableDelay = new Date(Date.now() - shortInterval);
    let query = initialRun ? findAllUnexpiredJobs() : Db.getGlobalDB()['jobs.Definition'].find().lessThan('nextExecution', acceptableDelay);

    return query.resultList();
  }).then(overdueJobs => {
    overdueJobs.forEach(job => {
      try {
        log('found overdue job, trying to reschedule: ' + job, 'warn');
        schedule(job, !initialRun);
      } catch (err) {
        Db.logError(err);
        log(err, 'warn');
      }
    });

    return checkForJobs();
  }).catch(e => {
    Db.logError(e);
    return true; // fallback to a short maintenance interval, if we had an error
  }).then(areThereActiveJobs => {
    return scheduleMaintenance(areThereActiveJobs);
  });
};

const newStatus = (localDb, job) => {
  const status = new localDb['jobs.Status']();
  status.cronjob = localDb['jobs.Definition'].ref(job.id);
  status.executedAt = new Date();
  return status;
};

const log = (message, loglevel) => {
  if (args.debug) {// Only used for local debugging
    if (Db.isInitialized())
      Db.getGlobalDB().log(loglevel || 'info', message);
    const pad = (number, digits)=> {
      let s = '' + number;
      while (s.length < digits) {
        s = '0' + s;
      }
      return s;
    };
    const now = new Date();
    let consoleMessage = pad(now.getHours(), 2) + ':' + pad(now.getMinutes(), 2) + ':' + pad(now.getSeconds(), 2) + '.' + pad(now.getMilliseconds(), 3) + ':\t' + (loglevel || 'info').toUpperCase() + '\t-\t' + message;
    if (loglevel === 'warn') {
      console.warn(consoleMessage);
    } else if (loglevel === 'error') {
      console.error(consoleMessage);
    } else {
      console.log(consoleMessage);
    }
  }
};

exports.handleJob = handleJob;
exports.maintenance = maintenance;