"use strict";

const vm = require('vm');
const Module = require('module');
//ensures that we always work with unix paths
const path = require('path').posix;

global.Abort = require('./abort');

const disabledModules = ['child_process', 'cluster', 'fs'];
let codes = {};
let codeModules = {};

const CodeModule = function CodeModule(id) {
  Object.defineProperties(this, {
    id: {
      value: id
    },
    filename: {
      value: id + '.js'
    }
  });
  this.exports = {};
};

CodeModule.load = function(id) {
  let codeModule = codeModules[id];

  if (codeModule)
    return codeModule;

  if (!codes[id])
    return null;

  const moduleParent = id.substring(0, id.lastIndexOf('/') + 1);

  function require(module) {
    if (module.indexOf('.') == 0) {
      module = path.join(moduleParent, module);
    }
    return CodeModule.require(module);
  }

  //prevent cycle dependency loading
  codeModule = codeModules[id] = new CodeModule(id);
  try {
    codes[id].apply(codeModule.exports, [codeModule.exports, require, codeModule, codeModule.filename, 'code']);
    return codeModule;
  } catch (e) {
    codeModules = {};
    throw e;
  }
};

CodeModule.require = function(module, defaultValue) {
  module = path.normalize(module);

  const codeModule = CodeModule.load(module);
  if (codeModule) {
    return codeModule.exports;
  }

  const index = module.indexOf('/');
  const nodeModule = index > -1 ? module.substring(0, index) : module;
  //nodeModule is an empty string if baqend code/handler is required
  if (nodeModule && disabledModules.indexOf(nodeModule) == -1)
    return require(module);

  if (defaultValue !== void 0)
    return defaultValue;

  throw new Error('The node module "' + module + '" is not available.');
};

CodeModule.compile = function(id, code) {
  const wrapper = Module.wrap(code);
  return vm.runInThisContext(wrapper, {
    filename: id + '.js'
  });
};

CodeModule.compileScript = function(id, code) {
  return new vm.Script(code, {
    filename: id + '.js'
  });
};

CodeModule.register = function(id, code) {
  codes[id] = CodeModule.compile(id, code);
  //force a reload of all modules
  codeModules = {};
};

CodeModule.clear = function() {
  codes = {};
  codeModules = {};
};

CodeModule.unregister = function(id) {
  codeModules = {};
  delete codes[id]
};

module.exports = CodeModule;
