"use strict";

const net = require('net');
const http = require('http');
const util = require('util');
const EventEmitter = require('events');
const url = require('url');

const STATUS_CODES = http.STATUS_CODES;

const PING = 0,
    PONG = 1,
    HEADER = 2,
    HEADER_FINAL = 3,
    DATA = 4,
    DATA_FINAL = 5,
    LOG = 6;


function Server(requestListener) {
  EventEmitter.call(this);

  if (requestListener) {
    this.addListener('request', requestListener);
  }

  this.pendingRequests = {};
}
util.inherits(Server, EventEmitter);

Server.prototype.listen = function(baqendUrl, callback) {
  const parsedUrl = url.parse(baqendUrl);
  const self = this;
  let restartWaitTime = 1;

  connect();

  function connect() {
    const client = net.connect({
      host: parsedUrl.hostname,
      port: parsedUrl.port,
      readable: true,
      writable: true
    });

    client.on('connect', self.connectionListener.bind(self, client));
    client.on('connect', callback);
    client.on('connect', client.removeListener.bind(client, 'error', reconnect));
    client.on('error', reconnect);
  }

  function reconnect(e) {
    restartWaitTime = restartWaitTime << 1;
    restartWaitTime = Math.min(restartWaitTime, 60);
    setTimeout(connect, restartWaitTime * 1000);
  }
};

Server.prototype.connectionListener = function(socket) {
  const self = this;

  let reqId = -1;
  let reqLength = 0;
  let reqType = -1;
  let incomingChunks = [];
  let incomingLength = 0;

  function dataListener(chunk) {
    incomingChunks.push(chunk);
    incomingLength += chunk.length;

    do {
      var loop = false;
      if (reqId == -1 && incomingLength >= 6) {
        let buffer = Buffer.concat(incomingChunks);

        //24bit length (16MB) + 8bit type + 16bit req id
        reqLength = buffer.readUIntBE(0, 3);
        reqType = buffer.readUIntBE(3, 1);
        reqId = buffer.readUIntBE(4, 2);

        incomingChunks = [buffer.slice(6)];
        incomingLength -= 6;
      }

      if (reqId != -1 && incomingLength >= reqLength) {
        let buffer = Buffer.concat(incomingChunks);
        let data = buffer.slice(0, reqLength);
        buffer = buffer.slice(reqLength);
        incomingChunks = [buffer];
        incomingLength = buffer.length;

        if (reqType == HEADER || reqType == HEADER_FINAL) {
          self.onHead(socket, reqId, data, reqType == HEADER_FINAL);
        } else if (reqType == DATA || reqType == DATA_FINAL) {
          self.onData(socket, reqId, data, reqType == DATA_FINAL)
        } else if (reqType == PING) {
          self.onPing(socket, reqId);
        } else {
          throw new Error('Protocol error: type=' + reqType + ', reqId=' + reqId + ', data=' + data);
        }

        reqId = -1;
        reqLength = 0;
        reqType = -1;
        loop = true;
      }
    } while (loop);
  }

  socket.on('data', dataListener);
  socket.on('close', function() {
    throw new Error('Connection closed unexpectedly');
  });
  socket.on('error', function(error) {
    throw new Error('Connection error occurred: ' + error);
  });
};

Server.prototype.onHead = function(socket, reqId, data, final) {
  const head = JSON.parse(data.toString());

  const req = new HttpRequest(socket, reqId, head);
  const res = new HttpResponse(req);

  if (!final) {
    this.pendingRequests[reqId] = req;
  }

  this.emit('request', req, res);

  if (final) {
    req.push(null);
  }
};

Server.prototype.onData = function(socket, reqId, data, final) {
  let req = this.pendingRequests[reqId];

  if (!req) {
    throw new Error('Protocol error: request not found reqId=' + reqId + ', data=' + data);
  }

  if (data.length > 0) {
    req.push(data);
  }

  if (final) {
    req.push(null);
    delete this.pendingRequests[reqId];
  }
};

Server.prototype.onPing = function(socket, reqId) {
  Server.write(socket, PONG, reqId);
};

Server.write = function(connection, type, reqId, data, encoding, callback) {
  let len;
  if (!data) {
    len = 0;
  } else if (typeof data == 'string') {
    len = Buffer.byteLength(data, encoding);
  } else {
    len = data.length;
  }

  //24bit length (16MB) + 8bit type + 16bit req id
  const head = new Buffer(6);
  head.writeUIntBE(len, 0, 3);
  head.writeUIntBE(type, 3, 1);
  head.writeUIntBE(reqId, 4, 2);

  let ret;
  if (len > 0) {
    connection.write(head);
    ret = connection.write(data, encoding, callback);
  } else {
    ret = connection.write(head, null, callback);
  }
  return ret;
};

var HttpRequest = function(socket, reqId, head) {
  http.IncomingMessage.call(this, socket);

  this.reqId = reqId;
  this.httpVersionMajor = 2;
  this.httpVersionMinor = 0;
  this.httpVersion = 2.0;
  this.complete = false;

  this.method = head.method;
  this.url = head.path;
  this.token = head.token;
  this.userId = head.userId;
  this.roleIds = head.roleIds;

  Object.keys(head.headers).forEach(name => {
    this._addHeaderLines([name, head.headers[name]], 1);
  });

  //make express body parser happy
  this._addHeaderLines(['transfer-encoding', 'chunked'], 1);
};

util.inherits(HttpRequest, http.IncomingMessage);

HttpRequest.prototype._read = function(n) {
  //nothing for now
};

function HttpResponse(req) {
  http.OutgoingMessage.call(this);

  this.socket = req.socket;
  this.reqId = req.reqId;
  this.connection = req.socket;
  this.emit('socket', req.socket);
  this.useChunkedEncodingByDefault = false;

  if (req.method === 'HEAD')
    this._hasBody = false;
}

util.inherits(HttpResponse, http.OutgoingMessage);

HttpResponse.prototype.statusCode = 200;
HttpResponse.prototype.statusMessage = undefined;

HttpResponse.prototype._implicitHeader = function() {
  this.writeHead(this.statusCode);
};

/** modifed version of Nodes ServerResponse.prototype.writeHead method */
HttpResponse.prototype.writeHead = function(statusCode, reason, obj) {
  let headers;

  if (typeof reason === 'string') {
    // writeHead(statusCode, reasonPhrase[, headers])
    this.statusMessage = reason;
  } else {
    // writeHead(statusCode[, headers])
    this.statusMessage = this.statusMessage || STATUS_CODES[statusCode] || 'unknown';
    obj = reason;
  }
  this.statusCode = statusCode;

  if (this._headers) {
    // Slow-case: when progressive API and header fields are passed.
    if (obj) {
      const keys = Object.keys(obj);
      for (let i = 0; i < keys.length; i++) {
        const k = keys[i];
        if (k) this.setHeader(k, obj[k]);
      }
    }
    // only progressive api is used
    headers = this._renderHeaders();
  } else {
    // only writeHead() called
    headers = obj;
  }

  if (statusCode === 204 || statusCode === 304 ||
      (100 <= statusCode && statusCode <= 199)) {
    this._hasBody = false;
  }

  this._header = JSON.stringify({
    status: this.statusCode,
    reason: this.statusMessage,
    headers: headers
  });
  this._headerSent = false;
};

HttpResponse.prototype.writeHeader = function() {
  this.writeHead.apply(this, arguments);
};

HttpResponse.prototype.write = function(chunk, encoding, callback) {
  if (typeof encoding === 'function') {
    callback = encoding;
    encoding = null;
  }

  if (this.finished) {
    const err = new Error('write after end');
    process.nextTick(writeAfterEndNT, this, err, callback);

    return true;
  }

  if (!this._header) {
    this._implicitHeader();
  }

  if (!this._hasBody) {
    debug('This type of response MUST NOT have a body. ' +
        'Ignoring write() calls.');
    return true;
  }

  if (typeof chunk !== 'string' && !(chunk instanceof Buffer)) {
    throw new TypeError('First argument must be a string or Buffer');
  }

  // If we get an empty string or buffer, then just do nothing, and
  // signal the user to keep writing.
  if (chunk.length === 0) return true;

  const ret = this._send(chunk, encoding, callback);

  return ret;
};

HttpResponse.prototype.end = function(data, encoding, callback) {
  if (typeof data === 'function') {
    callback = data;
    data = null;
  } else if (typeof encoding === 'function') {
    callback = encoding;
    encoding = null;
  }

  if (data && typeof data !== 'string' && !(data instanceof Buffer)) {
    throw new TypeError('First argument must be a string or Buffer');
  }

  if (this.finished) {
    return false;
  }

  const self = this;

  function finish() {
    self.emit('finish');
  }

  if (typeof callback === 'function')
    this.once('finish', callback);

  if (!this._header) {
    if (data) {
      if (typeof data === 'string')
        this._contentLength = Buffer.byteLength(data, encoding);
      else
        this._contentLength = data.length;
    } else {
      this._contentLength = 0;
    }
    this._implicitHeader();
  }

  if (data && !this._hasBody) {
    data = null;
  }

  //mark the end of
  this.finished = true;

  let ret;
  if (data) {
    // Normal body write.
    ret = this._send(data, encoding, finish);
  } else {
    // Flush headers
    ret = this._send('', 'binary', finish);
  }

  // There is the first message on the outgoing queue, and we've sent
  // everything to the socket.
  return ret;
};

HttpResponse.prototype._send = function(data, encoding, callback) {
  const final = this.finished;

  if (!this._headerSent) {
    if (final && data.length == 0) {
      const ret = this._writeRaw(HEADER_FINAL, this._header, 'utf8', callback);
      this._headerSent = true;
      return ret;
    } else {
      this._writeRaw(HEADER, this._header, 'utf8');
      this._headerSent = true;
    }
  }

  return this._writeRaw(final? DATA_FINAL: DATA, data, encoding, callback);
};

HttpResponse.prototype._writeRaw = function(type, data, encoding, callback) {
  const connection = this.connection;

  Server.write(connection, type, this.reqId, data, encoding, callback);
};

exports.Server = Server;
exports.HttpRequest = HttpRequest;
exports.HttpResponse = HttpResponse;